package seq.engine;

import java.util.Vector;

/**
 * @author kharaz + Elassad
 */

public final class Engine {


	public static final char[] allalstring
			= new char[] {'A','B','C', 'D', 'E', 'F', 'G', 'H', 'I', 'K','J','L',
						   'M', 'N','','','','X','U','O', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y','Z'};
	
	/** Working matrix */
	private Element[][] fMatrix;
	private Vector fTraceResult = new Vector();
	/** Reference of element has current max value. */
	private Element fMaxValue;
	
	/** Valid row characters on input sequence */
	private char[] fRowSeq;
	/** Valid column characters on input sequence */
	private char[] fColSeq;
	
	/** Buffer to store result. */
	private StringBuffer fRowBuffer;
	/** Buffer to store result. */
	private StringBuffer fColBuffer;
	/** Buffer to store result. */
	private StringBuffer fMatchLineBuffer;
	
	/** Matched element count. */
	private int fMatchCount;
	/** Mismatched element count. */
	private int fMismatchCount;
	/** Gap count */
	private int fGapCount;
	
	/** Match point. */
	private int fMatchWeight;
	/** Mismatch point. */
	private int fMismatchWeight;
	/** Gap penalty point. */
	private int fGapWeight;
	
	/** Valid character that recognizable to (Align) Engine. */
	private char[] fValidChars = allalstring;
	
	/** Case sentitive of not (default is not). */
	private boolean fCaseSentitive;	// false

	/**
	 * Returns result of best path tracing.
	 * 
	 * @return Coordinate of trace result.
	 */
	public int[][] getTraceResult() {
		int[][] result = new int[fTraceResult.size()][];
		for ( int i = 0; i < fTraceResult.size(); i++ ) {
			result[i] = (int[])fTraceResult.get(i);
		}
		return result;
	}
	
	/**
	 * Sets valid character that recognizable to (Align) Engine.
	 * 
	 * @param validChars Valid characters.
	 */
	public void setValidChars(String validChars) {
		fValidChars = validChars.toUpperCase().toCharArray();
	}
	
	/**
	 * Sets case sentitive of not.
	 * 
	 * @param caseSentitive Case sentitive or not.
	 */
	public void setCaseSentitive(boolean caseSentitive) {
		fCaseSentitive = caseSentitive;
	}
	
	/**
	 * Returns case sentitiviness.
	 * 
	 * @return Ture if case sentitive, otherwise false.
	 */
	public boolean isCaseSentitive() {
		return fCaseSentitive;
	}
	
	/**
	 * Valid character that recognizable to (Align) Engine.
	 * 
	 * @return Valid characters.
	 */
	public String getValidChars() {
		return new String(fValidChars);
	}
	
	/**
	 * Return founded matched elements count during alignment.
	 * 
	 * @return Matched elements count.
	 */
	public int getMatchCount() {
		return fMatchCount;
	}
	
	/**
	 * Return founded mismatched elements count during alignment.
	 * 
	 * @return Mismatched elements count.
	 */
	public int getMismatchCount() {
		return fMismatchCount;
	}
	
	/**
	 * Return founded gap count during alignment.
	 * 
	 * @return Gap count.
	 */
	public int getGapCount() {
		return fGapCount;
	}
	
	/**
	 * Return final score.
	 * @return Final score.
	 */
	public int getScore() {
		return ( fMatchCount * fMatchWeight )
				+ ( fMismatchCount * fMismatchWeight )
				+ ( fGapCount * fGapWeight );
	}
	
	/**
	 * Returns valid input sequence that recognizable to (Align) Engine.
	 *
	 * @return Valid input sequence.
	 */
	public char[] getValidInputSequenceA() {
		if ( fRowSeq == null ) {
			return null;
		} else {
			char[] seq = new char[fRowSeq.length];
			System.arraycopy(fRowSeq, 0, seq, 0, fRowSeq.length);
			return seq;
		}
	}
	
	/**
	 * Returns valid input sequence that recognizable to (Align) Engine.
	 *
	 * @return Valid input sequence.
	 */
	public char[] getValidInputSequenceB() {
		if ( fColSeq == null ) {
			return null;
		} else {
			char[] seq = new char[fColSeq.length];
			System.arraycopy(fColSeq, 0, seq, 0, fColSeq.length);
			return seq;
		}
	}
	
	/**
	 * Returns aligned characters of input sequence A.
	 * 
	 * @return Aligned characters of input sequence A.
	 */
	public String getResultSequenceA() {
		return fRowBuffer.toString();
	}
	
	/**
	 * Returns match/mismatch info line.
	 */
	public String getMatchLine() {
		return fMatchLineBuffer.toString();
	}
	
	/**
	 * Returns aligned characters of input sequence B.
	 * 
	 * @return Aligned characters of input sequence B.
	 */
	public String getResultSequenceB() {
		return fColBuffer.toString();
	}
	
	/**
	 * Returns working matrix.
	 * 
	 * @return Working matrix.
	 */
	public int[][] getMatrix() {
		int[][] matrix = null;
		if ( fMatrix != null ) {
			int row = fMatrix.length;
			int col = fMatrix[0].length;
			matrix = new int[row][col];
			for ( int r = 0; r < row; r++ ) {
				for ( int c = 0; c < col; c++ ) {
					matrix[r][c] = fMatrix[r][c].getValue();
				}
			}
		}
		return matrix;
	}
	
	/**
	 * Aligns two input sequences.
	 * 
	 * @param seqA Input sequence A.
	 * @param seqB Input sequence B.
	 * @param match Point added to matched elements.
	 * @param mismatch Point added to mismatched elements.
	 * @param gap Gap penalty point added to gap.
	 * @param global Global alignment or local alignment.
	 */
	public void process(String seqA, String seqB,
			int match, int mismatch, int gap, boolean global)
	{
		// initialinze data
		fMatchWeight = match;
		fMismatchWeight = mismatch;
		fGapWeight = gap;
		fRowSeq = getValidChars(seqA);
		fColSeq = getValidChars(seqB);
		// error check
		if (fRowSeq.length == 0 || fColSeq.length == 0) {
			throw new IllegalArgumentException();
		}
		fMatrix = new Element[fRowSeq.length + 1][fColSeq.length + 1];
		fRowBuffer = new StringBuffer();
		fColBuffer = new StringBuffer();
		fMatchLineBuffer = new StringBuffer();
		fMatchCount = 0;
		fMismatchCount = 0;
		fGapCount = 0;
		fTraceResult.clear();
		fMaxValue = null;
		// initializing end
		
		// fill matrix
		fillMatrix(global);
		// trace
		if ( global ) {
			trace( fMatrix[fRowSeq.length][fColSeq.length], true );
			fTraceResult.add( new int[] {	fMatrix[0][0].getColumn(),
											fMatrix[0][0].getRow()} );
		} else {
			trace( fMaxValue, false);
		}
		
		// result process
		fRowBuffer.reverse();
		fColBuffer.reverse();
		fMatchLineBuffer.reverse();
	}
	
	/**
	 * Traces best path after matric has filled.
	 * 
	 * @param source Start point to trace.
	 * @param global Global alignment of local alignment.
	 */
	private void trace(Element source, boolean global) {
		if ( source == null ) {
			return;
		}
		if ( !global && source.getValue() == 0 ) {
			fTraceResult.add(new int[] {source.getColumn(), source.getRow()});
			return;
		}
		
		if ( source.isFrom(Element.NORTH_WEST) ) {
			if ( source.isMatched() ) {
				fMatchCount++;
				fMatchLineBuffer.append('|');
			} else {
				fMismatchCount++;
				fMatchLineBuffer.append(' ');
			}
			fRowBuffer.append( fRowSeq[source.getRow() - 1] );
			fColBuffer.append( fColSeq[source.getColumn() - 1] );
			fTraceResult.add(new int[] {source.getColumn(), source.getRow()});
			trace(source.getFrom(Element.NORTH_WEST), global);
		} else if ( source.isFrom(Element.NORTH) ) {
			fGapCount++;
			fRowBuffer.append( fRowSeq[source.getRow() - 1] );
			fColBuffer.append( ' ' );
			fMatchLineBuffer.append( ' ' );
			fTraceResult.add(new int[] {source.getColumn(), source.getRow()});
			trace(source.getFrom(Element.NORTH), global);
		} else if ( source.isFrom(Element.WEST) ) {
			fGapCount++;
			fRowBuffer.append( ' ' );
			fColBuffer.append( fColSeq[source.getColumn() - 1] );
			fMatchLineBuffer.append( ' ' );
			fTraceResult.add(new int[] {source.getColumn(), source.getRow()});
			trace(source.getFrom(Element.WEST), global);
		}
	}
	
	/**
	 * Calculates matrix.
	 * 
	 * @param global Global alignment or local alignment.
	 */
	private void fillMatrix(boolean global) {
		
		int hiddenValue = Integer.MIN_VALUE;
		if ( !global ) {
			hiddenValue = 0;
		}
		
		// filling element at (0, 0)
		fMatrix[0][0] = new Element(0, 0, 0);
		
		// filling first column
		int gapWeight = fGapWeight;
		if ( !global ) {
			gapWeight = 0;
		}
		for ( int row = 1; row <= fRowSeq.length; row++ ) {
			Element element = new Element(row, 0, row * gapWeight);
			element.setFrom(Element.NORTH, fMatrix[row - 1][0]);
			fMatrix[row][0] = element;
		}
		// filling first row
		for ( int col = 1; col <= fColSeq.length; col++ ) {
			Element element = new Element(0, col, col * gapWeight);
			element.setFrom(Element.WEST, fMatrix[0][col - 1]);
			fMatrix[0][col] = element;
		}
		
		// filling others
		int maxValue = Integer.MIN_VALUE;
		for ( int row = 1; row <= fRowSeq.length; row++ ) {
			for ( int col = 1; col <= fColSeq.length; col++ ) {
				// create new Element
				Element element = new Element(row, col);
				
				// match or mismatch
				element.setMatched( fRowSeq[row - 1] == fColSeq[col - 1] );
				int matchValue = element.isMatched()
						? fMatrix[row - 1][col - 1].getValue() + fMatchWeight
						: fMatrix[row - 1][col - 1].getValue() + fMismatchWeight;
				// row skip
				int rowSkipValue
						= fMatrix[row - 1][col].getValue() + fGapWeight;
				// column skip
				int columnSkipValue
						= fMatrix[row][col - 1].getValue() + fGapWeight;
				// get max value
				int max = getMax(	matchValue, rowSkipValue,
									columnSkipValue, hiddenValue );
				
				// set value
				element.setValue(max);
				// set link
				if ( matchValue == max ) {
					element.setFrom(Element.NORTH_WEST,
									fMatrix[row - 1][col - 1]);
				}
				if ( rowSkipValue == max ) {
					element.setFrom(Element.NORTH, fMatrix[row - 1][col]);
				}
				if ( columnSkipValue == max ) {
					element.setFrom(Element.WEST, fMatrix[row][col - 1]);
				}
				
				fMatrix[row][col] = element;

				// rememver max value
				if ( max > maxValue ) {
					maxValue = max;
					fMaxValue = element;
				}
			}
		}
	}
	
	/**
	 * Returns maximum value of specified four values.
	 * 
	 * @param value1 Possible value 1.
	 * @param value2 Possible value 2.
	 * @param value3 Possible value 3.
	 * @param value4 Possible value 4.
	 * @return Max value
	 */
	private static int getMax(int value1, int value2, int value3,
			int value4)
	{
		int maxValue = (value1 > value2) ? value1 : value2;
		maxValue = (maxValue > value3) ? maxValue : value3;
		maxValue = (maxValue > value4) ? maxValue : value4;
		return maxValue;
	}
	
	/**
	 * Extracts valid characters in raw input string.
	 * 
	 * @param seq Input sequence as String.
	 * @return valid characters as char array.
	 */
	private char[] getValidChars(String seq) {
		char[] seqChars;
		
		if ( fCaseSentitive ) {
			// case sentitive alignment
			seqChars = seq.toCharArray();
		} else {
			seqChars = seq.toUpperCase().toCharArray();
		}
		
		// Extracts valid character in raw input.
		StringBuffer seqBuffer = new StringBuffer();
		for ( int i = 0; i < seqChars.length; i++ ) {
			for ( int j = 0; j < fValidChars.length; j++ ) {
				if ( seqChars[i] == fValidChars[j] ) {
					seqBuffer.append(seqChars[i]);
				}
			}
		}
		
		// Returns valid caharacters.
		return seqBuffer.toString().toCharArray();
	}
}
